// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#pragma once
#include <array>
#include <cmath>
#include <string>

/*! An enumeration of color spaces with CIE XYZ as underlying color matching
	functions.*/
enum class TristimulusColorSpace{
	xyz,
	sRGB
};

/*! An enumeration of color spaces with Fourier basis functions as underlying
	color matching functions.*/
enum class FourierColorSpace : int{
	/*! Three unmodified Fourier coefficients. If the first Fourier coefficient
		is complex, its real part is in dimension 1 and its complex part in
		dimension 2.*/
	plain = 0,
	/*! Only meaningful for mirrored and warped Fourier coefficients.
		\see convertMirroredWarpedFourierToFourierXYZ() */
	fourierXYZ = 1,
	/*! Only meaningful for mirrored and warped Fourier coefficients.
		\see convertMirroredWarpedFourierToFourierSRGB() */
	fourierSRGB = 2
};

/*! Returns a name for the given tristimulus color space to use in file
	names.*/
inline std::string getColorSpaceName(TristimulusColorSpace space) {
	switch(space) {
	case TristimulusColorSpace::xyz:
		return "XYZ";
	case TristimulusColorSpace::sRGB:
		return "sRGB";
	default:
		return "unknown";
	}
}


/*! Returns a name for the given Fourier color space to use in file names.*/
inline std::string getColorSpaceName(FourierColorSpace space) {
	switch(space) {
	case FourierColorSpace::plain:
		return "Fourier";
	case FourierColorSpace::fourierXYZ:
		return "FourierXYZ";
	case FourierColorSpace::fourierSRGB:
		return "FourierSRGB";
	default:
		return "unknown";
	}
}


/*! This utility function evaluates the non-linearity for
	convertXYZToCIELab().*/
inline float nonLinearityXYZToCIELAB(float t) {
	const float delta = 6.0f / 29.0f;
	return (t > delta*delta*delta) ? std::pow(t, 1.0f / 3.0f) : (t / (3.0f*delta*delta) + 4.0f / 29.0f);
}


/*! Maps an XYZ triple to a triple in the CIELAB color space, assuming either
	an illuminant E whitepoint or a D65 whitepoint. For triples with negative
	luminance a vector of -1.0f is returned.*/
inline std::array<float, 3> convertXYZToCIELAB(const std::array<float, 3>& rXYZ, bool useD65) {
	if (rXYZ[1] < 0.0f) {
		return { -1.0f, -1.0f, -1.0f };
	}
	float invWhitePointX = useD65 ? (1.0f / 10043.8672f) : (1.0f / 106.865685f);
	float invWhitePointY = useD65 ? (1.0f / 10567.2900f) : (1.0f / 106.856987f);
	float invWhitePointZ = useD65 ? (1.0f / 11507.4521f) : (1.0f / 106.893394f);
	float nonLinearX = nonLinearityXYZToCIELAB(rXYZ[0] * invWhitePointX);
	float nonLinearY = nonLinearityXYZToCIELAB(rXYZ[1] * invWhitePointY);
	float nonLinearZ = nonLinearityXYZToCIELAB(rXYZ[2] * invWhitePointZ);
	return {
		116.0f * nonLinearY - 16.0f,
		500.0f * (nonLinearX - nonLinearY),
		200.0f * (nonLinearY - nonLinearZ)
	};
}


/*! This utility function evaluates the non-linearity for
	convertSRGBToXYZ(). It does not include any clamping but instead continues
	the linear gradient of 1.0f/12.92f for negative values.*/
inline float nonLinearitySRGBToXYZ(float t) {
	return (t > 0.04045f) ? std::pow((t + 0.055f) / 1.055f, 2.4f) : (t / 12.92f);
}


/*! Implements the inverse of nonLinearitySRGBToXYZ().*/
inline float nonLinearityXYZToSRGB(float t) {
	return (t > 0.00313080495f) ? (1.055f * std::pow(t, 1.0f / 2.4f) - 0.055f) : (12.92f * t);
}


/*! This function converts an sRGB triple to an XYZ triple. The normalization
	is done such that the input triple (1,1,1) corresponds to the XYZ for a
	reflectance spectrum that is constant one, lit by illuminant E or
	illuminant D65. There is no clampling involved, so except for rounding
	errors, the operation is bijective.
	\sa nonLinearitySRGBToXYZ()
	\sa convertXYZToSRGB() */
inline std::array<float, 3> convertSRGBToXYZ(const std::array<float, 3>& rSRGB, bool useD65) {
	std::array<float, 3> linearSRGB = {
		nonLinearitySRGBToXYZ(rSRGB[0]),
		nonLinearitySRGBToXYZ(rSRGB[1]),
		nonLinearitySRGBToXYZ(rSRGB[2])
	};
	if (useD65)
	{
		std::array<float, 3> xyz = {
			0.4124f*linearSRGB[0] + 0.3576f*linearSRGB[1] + 0.1805f*linearSRGB[2],
			0.2126f*linearSRGB[0] + 0.7152f*linearSRGB[1] + 0.0722f*linearSRGB[2],
			0.0193f*linearSRGB[0] + 0.1192f*linearSRGB[1] + 0.9505f*linearSRGB[2]
		};
		const float denormalization = 10567.2949f;
		return { denormalization * xyz[0], denormalization * xyz[1], denormalization * xyz[2] };
	}
	else
	{
		std::array<float, 3> xyz = {
			0.49685929f*linearSRGB[0] + 0.33909388f*linearSRGB[1] + 0.16404638f*linearSRGB[2],
			0.25619309f*linearSRGB[0] + 0.67818808f*linearSRGB[1] + 0.06561853f*linearSRGB[2],
			0.02329024f*linearSRGB[0] + 0.11303116f*linearSRGB[1] + 0.86397868f*linearSRGB[2]
		};
		const float denormalization = 106.85706329345703f;
		return { denormalization * xyz[0], denormalization * xyz[1], denormalization * xyz[2] };
	}
}


/*! Implements the reverse of convertSRGBToXYZ().*/
inline std::array<float, 3> convertXYZToSRGB(const std::array<float, 3>& rXYZ, bool useD65) {
	std::array<float, 3> linearSRGB;
	if (useD65)
	{
		linearSRGB[0] = +3.06665567e-4f*rXYZ[0] - 1.45468446e-4f*rXYZ[1] - 4.71860210e-5f*rXYZ[2];
		linearSRGB[1] = -9.16914616e-5f*rXYZ[0] + 1.77505793e-4f*rXYZ[1] + 3.92886961e-6f*rXYZ[2];
		linearSRGB[2] = +5.27193771e-6f*rXYZ[0] - 1.93068380e-5f*rXYZ[1] + 1.00025215e-4f*rXYZ[2];
	}
	else
	{
		linearSRGB[0] = +2.51737127e-2f*rXYZ[0] - 1.19413723e-2f*rXYZ[1] - 3.87287466e-3f*rXYZ[2];
		linearSRGB[1] = -9.56506730e-3f*rXYZ[0] + 1.85131514e-2f*rXYZ[1] + 4.10089849e-4f*rXYZ[2];
		linearSRGB[2] = +5.72755850e-4f*rXYZ[0] - 2.10010454e-3f*rXYZ[1] + 1.08823785e-2f*rXYZ[2];
	}
	return {
		nonLinearityXYZToSRGB(linearSRGB[0]),
		nonLinearityXYZToSRGB(linearSRGB[1]),
		nonLinearityXYZToSRGB(linearSRGB[2])
	};
}


/*! This function converts a triple of Fourier coefficients generated using
	mirroring and warping to a color space that resembles linear XYZ (with
	respect to illuminant E). This is accomplished by a simple linear
	transform. Note that the warp is designed to ensure that such a linear
	transform can get reasonably close to XYZ.*/
inline std::array<float, 3> convertMirroredWarpedFourierToFourierXYZ(const std::array<float, 3>& rFourier) {
	std::array<float, 3> fourierXYZ = {
		0.1566469241f*rFourier[0] + 0.1476465788f*rFourier[1] + 0.1010779268f*rFourier[2],
		0.1562831028f*rFourier[0] + 0.1066063399f*rFourier[1] - 0.0885138259f*rFourier[2],
		0.1592744056f*rFourier[0] - 0.2523955038f*rFourier[1] + 0.1509378426f*rFourier[2],
	};
	// Computed by computeFourierXYZNormalizationFactor()
	const float normalizationFactor = 683.739868f;
	return { normalizationFactor*fourierXYZ[0], normalizationFactor*fourierXYZ[1], normalizationFactor*fourierXYZ[2] };
}


/*! Implements the inverse of convertMirroredWarpedFourierToFourierXYZ().*/
inline std::array<float, 3> convertFourierXYZToMirroredWarpedFourier(const std::array<float, 3>& rFourierXYZ) {
	// Computed by computeFourierXYZNormalizationFactor()
	const float denormalizationFactor = 1.0f / 683.739868f;
	std::array<float, 3> fourierXYZ = { denormalizationFactor*rFourierXYZ[0], denormalizationFactor*rFourierXYZ[1], denormalizationFactor*rFourierXYZ[2] };
	return {
		0.5103082995f*fourierXYZ[0] + 3.9028729757f*fourierXYZ[1] + 1.9470088419f*fourierXYZ[2],
		3.0773362531f*fourierXYZ[0] - 0.6160729511f*fourierXYZ[1] - 2.4220681578f*fourierXYZ[2],
		4.6073719528f*fourierXYZ[0] - 5.1486214645f*fourierXYZ[1] + 0.5205600513f*fourierXYZ[2],
	};
}


/*! This function converts a triple of Fourier coefficients generated using
	mirroring and warping to a color space that resembles sRGB (with respect to
	illuminant E). This is accomplished by a simple linear transform followed
	by the non-linearity of sRGB. Note that the warp is designed to ensure that
	such a linear transform can get reasonably close to tristimulus color
	spaces. Fourier sRGB is defined in such a way that each sRGB triple in
	[0,1] can be represented by a Fourier sRGB triple in [0,1].
	\note Converting Fourier XYZ to sRGB using convertXYZToSRGB() does _not_
		give you Fourier sRGB (or sRGB). Do not do that.*/
inline std::array<float, 3> convertMirroredWarpedFourierToFourierSRGB(const std::array<float, 3>& rFourier) {
	std::array<float, 3> linearFourierSRGB = {
		0.9983434273f*rFourier[0] + 2.2557328657f*rFourier[1] + 2.0053616188f*rFourier[2],
		0.9983812507f*rFourier[0] + 0.1131528687f*rFourier[1] - 1.6007375361f*rFourier[2],
		0.9974500349f*rFourier[0] - 1.8951714334f*rFourier[1] + 1.0141189901f*rFourier[2],
	};
	return { nonLinearityXYZToSRGB(linearFourierSRGB[0]), nonLinearityXYZToSRGB(linearFourierSRGB[1]), nonLinearityXYZToSRGB(linearFourierSRGB[2]) };
}


/*! Implements the inverse of convertMirroredWarpedFourierToFourierSRGB().*/
inline std::array<float, 3> convertFourierSRGBToMirroredWarpedFourier(const std::array<float, 3>& rFourierSRGB) {
	std::array<float, 3> linearFourierSRGB = { nonLinearitySRGBToXYZ(rFourierSRGB[0]), nonLinearitySRGBToXYZ(rFourierSRGB[1]), nonLinearitySRGBToXYZ(rFourierSRGB[2]) };
	return {
		0.2276800312f*linearFourierSRGB[0] + 0.4748793279f*linearFourierSRGB[1] + 0.2993498517f*linearFourierSRGB[2],
		0.2035160895f*linearFourierSRGB[0] + 0.0770505049f*linearFourierSRGB[1] - 0.2808208130f*linearFourierSRGB[2],
		0.1563903501f*linearFourierSRGB[0] - 0.3230828824f*linearFourierSRGB[1] + 0.1668540861f*linearFourierSRGB[2],
	};
}


/*! This function converts from an arbitrary Fourier color space to plain
	Fourier.
	\param source A color in inputSpace.
	\return A color in FourierColorSpace::plain.*/
inline std::array<float, 3> convertAnyFourierToMirroredWarpedFourier(std::array<float, 3> input, FourierColorSpace inputSpace) {
	switch (inputSpace) {
	case FourierColorSpace::fourierXYZ:
		return convertFourierXYZToMirroredWarpedFourier(input);
	case FourierColorSpace::fourierSRGB:
		return convertFourierSRGBToMirroredWarpedFourier(input);
	default:
		return input;
	}
}


/*! This function converts between arbitrary Fourier color spaces.
	\param source A color in inputSpace.
	\return A color in outputSpace.*/
inline std::array<float, 3> convertAnyFourierToAnyFourier(std::array<float, 3> input, FourierColorSpace inputSpace, FourierColorSpace outputSpace) {
	int iCase = static_cast<int>(inputSpace) * 3 + static_cast<int>(outputSpace);
	switch (iCase) {
	case static_cast<int>(FourierColorSpace::plain)*3 + static_cast<int>(FourierColorSpace::fourierXYZ):
		return convertMirroredWarpedFourierToFourierXYZ(input);
	case static_cast<int>(FourierColorSpace::plain)*3 + static_cast<int>(FourierColorSpace::fourierSRGB):
		return convertMirroredWarpedFourierToFourierSRGB(input);
	case static_cast<int>(FourierColorSpace::fourierXYZ)*3 + static_cast<int>(FourierColorSpace::plain):
		return convertFourierXYZToMirroredWarpedFourier(input);
	case static_cast<int>(FourierColorSpace::fourierXYZ)*3 + static_cast<int>(FourierColorSpace::fourierSRGB):
		return convertMirroredWarpedFourierToFourierSRGB(convertFourierXYZToMirroredWarpedFourier(input));
	case static_cast<int>(FourierColorSpace::fourierSRGB)*3 + static_cast<int>(FourierColorSpace::plain):
		return convertFourierSRGBToMirroredWarpedFourier(input);
	case static_cast<int>(FourierColorSpace::fourierSRGB)*3 + static_cast<int>(FourierColorSpace::fourierXYZ):
		return convertMirroredWarpedFourierToFourierXYZ(convertFourierSRGBToMirroredWarpedFourier(input));
	default:
		return input;
	}
}
